import csv
import glob
import os
from collections import OrderedDict
from typing import Dict, Tuple, List

import pandas as pd

from config import TEMPORARY_OUTPUT_PATH
from study2.s2_config import IGNORE_FILE_NR, COLOR_MAP, DATA_SET_INPUT_PATH, STUDY_INPUT_PATH, QUESTIONNAIRE_INPUT_PATH, VIS_MAP
from utils import data_string_converter


def get_visualization_data(reload_csv: bool = False) -> pd.DataFrame:
    os.makedirs(TEMPORARY_OUTPUT_PATH, exist_ok=True)
    file_path = f"{TEMPORARY_OUTPUT_PATH}/S2-visualization-data.csv"
    if os.path.isfile(file_path) and not reload_csv:
        return pd.read_csv(file_path)

    output = None

    for fp in glob.glob(f"{DATA_SET_INPUT_PATH}/*.csv"):
        file_name = fp.split('\\')[-1]
        if file_name[0] == "T":
            continue

        file_nr = int(file_name[:3]) - 1
        block_nr = VIS_MAP[file_nr]["block"]
        vis_nr = VIS_MAP[file_nr]["vis"]

        df = pd.read_csv(fp, sep=";", skipinitialspace=True)
        df["blockType"] = block_nr
        df["visNr"] = vis_nr

        if output is None:
            output = df
        else:
            output = output.append(df, ignore_index=True)

    output.to_csv(file_path)
    return output


def get_answer_data(reload_csv: bool = False) -> pd.DataFrame:
    os.makedirs(TEMPORARY_OUTPUT_PATH, exist_ok=True)
    file_path = f"{TEMPORARY_OUTPUT_PATH}/S2-user-answer-data.csv"
    if os.path.isfile(file_path) and not reload_csv:
        df = pd.read_csv(file_path)
        df["answer"] = df.apply(lambda row: data_string_converter(row["answer"]), axis=1)
        return df

    output = None
    for dic in [d[0] for d in os.walk(STUDY_INPUT_PATH)]:
        for fp in glob.glob(f"{dic}/AnswerLog-*.csv"):
            file_name = fp.split('\\')[-1]
            _, file_nr, background = file_name.split('-')
            file_nr = int(file_nr)
            if int(file_nr) in IGNORE_FILE_NR:
                continue

            df = pd.read_csv(fp, sep=";", skipinitialspace=True)
            df = df[~df.blockNr.isin([-1, -2, -3])].reset_index(drop=True)
            df["fileNr"] = file_nr
            df["background"] = background[1:4]

            def apply_block_order(row: pd.Series, with_offset: bool = True):
                offset = 0
                if with_offset:
                    if file_nr % 2 == 1:
                        if row["background"] == "BG3":
                            offset = 0
                        else:
                            offset = 12
                    else:
                        if row["background"] == "BG3":
                            offset = 12
                        else:
                            offset = 0

                index = (int(row["sessionOrder"] / 4) + offset) % 24 + 1
                return index

            df["sessionOrder"] = [i for i in range(0, df.shape[0])]
            df["backgroundOrder"] = df.apply(lambda row: apply_block_order(row, False), axis=1)
            df["sessionOrder"] = df.apply(lambda row: apply_block_order(row), axis=1)

            if output is None:
                output = df
            else:
                output = output.append(df, ignore_index=True)

    output.reset_index()
    output = output.drop("day", axis=1)

    def convert_answers(row: pd.Series):
        if "," in row["answer"]:
            answer = [COLOR_MAP.index(s) for s in row["answer"].split(',') if len(s) is 1]
        elif "-" in row["answer"]:
            answer = tuple(row["answer"].split('-'))
            answer = int(answer[0]), int(answer[1])
        elif row["answer"] in ["b", "g", "w", "r"]:
            answer = COLOR_MAP.index(row["answer"])
        elif row["answer"] in ["j", "n"]:
            answer = True if row["answer"] == "j" else False
        else:
            answer = int(row["answer"])

        # Special checks for the current study.
        if row["blockType"] in [1, 2] and row["questionNr"] is 1:
            if type(answer) is not list:
                answer = [answer]
        elif row["blockType"] is 2 and row["questionNr"] is 4:
            if type(answer) is not tuple:
                answer = (-1, -1)

        return answer

    output["blockType"] = output.apply(lambda row: int((row.blockNr - 1) / 8) + 1, axis=1)
    output["visNr"] = output.apply(lambda row: row.blockNr % 8 if row.blockNr not in [8, 16, 24] else 8, axis=1)
    output["answer"] = output.apply(lambda row: convert_answers(row), axis=1)
    output = output.reset_index()

    output.to_csv(file_path)
    return output


def get_time_data() -> Dict[int, Dict[Tuple[int, int, str], List[List[str]]]]:
    """
    :return: Dict with filenr as key and another dict with (block, question, background) as key
    """
    output = {}

    for dic in [d[0] for d in os.walk(STUDY_INPUT_PATH)]:
        for fp in glob.glob(f"{dic}/ClientLog(AR)-*.csv"):
            file_name = fp.split('\\')[-1]
            _, file_nr, background = file_name.split('-')
            file_nr = int(file_nr)
            if file_nr in IGNORE_FILE_NR:
                continue

            if file_nr not in output:
                output[file_nr] = OrderedDict()

            with open(fp, newline="") as csv_file:
                reader = list(csv.reader(csv_file, delimiter=';'))
                reader.pop(0)
                reader.pop(0)
                while "StopApp" not in reader[0][3]:
                    index = 0
                    for i, row in enumerate(reader):
                        if i == 0:
                            continue
                        if "NextQuestion" in row[3] or "StopApp" in row[3] or "NextBlock" in row[3]:
                            index = i
                            break

                    b = int(reader[0][4])
                    q = int(reader[0][5])
                    if b <= 0:
                        reader.pop(0)
                        continue

                    q_desc = (b, q if q != 0 else 1, background[1:4])
                    output[file_nr][q_desc] = []
                    for i in range(0, index):
                        output[file_nr][q_desc].append(reader[0])
                        reader.pop(0)
                    output[file_nr][q_desc].append(reader[0])

                    if "NextBlock" in reader[0][3]:
                        reader.pop(0)

    return output


def get_questionnaire_data():
    if os.path.isfile(QUESTIONNAIRE_INPUT_PATH):
        df = pd.read_csv(QUESTIONNAIRE_INPUT_PATH)
        return df

    return None

