import datetime
import itertools
import json
import os
from typing import Dict, List

import pandas as pd
import pingouin as pg

from study2.base_data_processing import get_visualization_data, get_answer_data, get_time_data
from config import TEMPORARY_OUTPUT_PATH
from study2.s2_config import QuestionDesc, QUESTION_TYPES_CALCULATED, QUESTION_TYPE_MAP, STUDY_ORDER_INPUT_PATH
from study2.s2_utils import convert_order_values
from utils import data_string_converter, find_indices


def calc_values(reload_csv: bool = False) -> pd.DataFrame:
    file_path = f"{TEMPORARY_OUTPUT_PATH}/S2-calcualted-data.csv"
    if os.path.isfile(file_path) and not reload_csv:
        df = pd.read_csv(file_path)
        df["correctAnswer"] = df.apply(lambda row: data_string_converter(row["correctAnswer"]), axis=1)
        df["answer"] = df.apply(lambda row: data_string_converter(row["answer"]), axis=1)
        df["calculatedValue"] = df.apply(lambda row: data_string_converter(row["calculatedValue"]), axis=1)
        df["trimmedValue"] = df.apply(lambda row: data_string_converter(row["trimmedValue"]), axis=1)
        return df

    vis_data = get_visualization_data()
    answer_data = get_answer_data(True)

    solution = {}
    solution.update(calc_values_block1(vis_data, answer_data))
    solution.update(calc_values_block2(vis_data, answer_data))
    solution.update(calc_values_block3(vis_data, answer_data))

    with open(STUDY_ORDER_INPUT_PATH, newline="") as json_file:
        study_order = json.load(json_file)

        tmp_results = {"fileNr": [], "background": [], "focus": [], "sessionOrder": [], "backgroundOrder": [], "blockType": [], "visNr": [],
                       "questionType": [], "questionNr": [], "note": [],
                       "correctAnswer": [], "answer": [], "calculatedValue": [], "valueType": []}
        for q_desc, value_list in solution.items():
            for values in value_list:
                tmp_results["fileNr"].append(q_desc.fileNr)
                tmp_results["background"].append(q_desc.background)
                tmp_results["focus"].append(study_order[f"{q_desc.fileNr}-{q_desc.background}"][f"{q_desc.blockNr:03d}"]["withFocusChange"])
                tmp_results["sessionOrder"].append(q_desc.sessionOrder)
                tmp_results["backgroundOrder"].append(q_desc.backgroundOrder)
                tmp_results["blockType"].append(q_desc.blockType)
                tmp_results["visNr"].append(q_desc.visNr)
                tmp_results["questionType"].append(QUESTION_TYPE_MAP[q_desc.blockType][values["questionNr"]] if values["note"] == "-" else values["questionType"])
                tmp_results["questionNr"].append(values["questionNr"])
                tmp_results["note"].append(values["note"])
                tmp_results["correctAnswer"].append(values["correctAnswer"])
                tmp_results["answer"].append(values["answer"] if "answer" in values else "")
                tmp_results["calculatedValue"].append(values["calculatedValue"])
                tmp_results["valueType"].append(values["valueType"])

    # Trim data
    def apply_trim_values(row, mean_df, error_df):
        if row["valueType"] != "NUMERIC":
            return row["calculatedValue"]

        new_value = float(mean_df[mean_df.questionType == row["questionType"]]["calculatedValue"] +
                          2 * error_df[error_df.questionType == row["questionType"]]["calculatedValue"])
        return row["calculatedValue"] if row["calculatedValue"] <= new_value else new_value

    solution_df = pd.DataFrame(tmp_results)
    solution_df["calculatedValue"] = solution_df["calculatedValue"].astype("float64")
    value_mean = solution_df.groupby(by=["questionType"]).mean().reset_index()
    value_std = solution_df.groupby(by=["questionType"]).std().reset_index()
    solution_df["trimmedValue"] = solution_df.apply(lambda row: apply_trim_values(row, value_mean, value_std), axis=1)
    solution_df = convert_order_values(solution_df, "sessionOrder")
    solution_df = convert_order_values(solution_df, "backgroundOrder")

    solution_df.to_csv(file_path)
    return solution_df


def calc_values_block1(vis_data: pd.DataFrame, answer_data: pd.DataFrame) -> Dict[QuestionDesc, List[Dict]]:
    question_values1 = [
        [(5, 10), "h"],
        [(15, 20), "h"],
        [(5, 10), "h"],
        [(10, 15), "h"],
        [(20, 25), "h"],
        [(40, 45), "l"],
        [(25, 30), "h"],
        [(35, 42), "l"]
    ]

    solution = {}
    for i, ((min_range, max_range), extremum) in enumerate(question_values1):
        vis_df = vis_data[(vis_data.blockType == 1) & (vis_data.visNr == i + 1)]
        vis_df_1 = vis_df[(vis_df.Y > min_range) & (vis_df.Y < max_range)]
        answer_1 = [l for l in range(0, 4) if vis_df_1[vis_df_1.Line == l].count()["Line"] >= 3]

        for file_nr in answer_data["fileNr"].unique():
            results = []
            answer_df = answer_data[(answer_data.blockType == 1) & (answer_data.visNr == i + 1) & (answer_data.fileNr == file_nr)]
            correct_answers = {
                1: list(answer_df[answer_df.questionNr == 1]["answer"])[0],
                2: list(answer_df[answer_df.questionNr == 2]["answer"])[0],
                3: list(answer_df[answer_df.questionNr == 3]["answer"])[0],
                4: list(answer_df[answer_df.questionNr == 4]["answer"])[0],
            }

            # Question 1
            results.append({
                "questionNr": 1,
                "note": "-",
                "correctAnswer": answer_1,
                "answer": correct_answers[1],
                "calculatedValue": set(answer_1) == set(correct_answers[1]),
                "valueType": "BOOL"
            })
            results.append({
                "questionNr": 1,
                "note": "# wrong",
                "questionType": QUESTION_TYPES_CALCULATED[2],
                "correctAnswer": 0,
                "calculatedValue": len([v for v in correct_answers[1] if v not in answer_1]),
                "valueType": "NUMERIC"
            })
            # Question 2 + 3
            vis_df_2 = vis_df[vis_df.Line.isin(correct_answers[1])]
            if extremum == "h":
                answer_3 = vis_df_2.max()["Y"]
            else:
                answer_3 = vis_df_2.min()["Y"]
            answer_2 = list(vis_df_2[vis_df_2.Y == answer_3]["X"])

            results.append({
                "questionNr": 2,
                "note": "-",
                "correctAnswer": answer_2,
                "answer": correct_answers[2],
                "calculatedValue": correct_answers[2] in answer_2,
                "valueType": "BOOL"
            })
            results.append({
                "questionNr": 3,
                "note": "-",
                "correctAnswer": answer_3,
                "answer": correct_answers[3],
                "calculatedValue": abs(answer_3 - correct_answers[3]),
                "valueType": "NUMERIC"
            })
            # Question 4
            answer_4 = list(vis_df_2[vis_df_2.Y == answer_3]["Line"])
            vis_df_4 = vis_df[vis_df.Line == answer_4[0]]
            vis_df_4 = vis_df_4.groupby(by="Line").mean().reset_index()
            answer_4 = list(vis_df_4["Y"])[0]
            results.append({
                "questionNr": 4,
                "note": "-",
                "correctAnswer": answer_4,
                "answer": correct_answers[4],
                "calculatedValue": abs(answer_4 - correct_answers[4]),
                "valueType": "NUMERIC"
            })

            q_desc = QuestionDesc(file_nr, list(answer_df["background"])[0], list(answer_df["backgroundOrder"])[0], list(answer_df["sessionOrder"])[0], list(answer_df["blockNr"])[0], 1, i + 1)
            solution[q_desc] = results

    return solution


def calc_values_block2(vis_data: pd.DataFrame, answer_data: pd.DataFrame) -> Dict[QuestionDesc, List[Dict]]:
    question_values2 = [
        ["h", 24],
        ["h", 17],
        ["h", 31],
        ["h", 16],
        ["l", 19],
        ["l", 16],
        ["l", 7],
        ["l", 12],
    ]

    solution = {}
    for i, (sort, x_value) in enumerate(question_values2):
        vis_df = vis_data[(vis_data.blockType == 2) & (vis_data.visNr == i + 1)]
        if sort == "h":
            vis_df_1 = vis_df.groupby(by=["Line"]).max().reset_index()
            vis_df_1 = vis_df_1.sort_values(by=["Y"], ascending=False)
        else:
            vis_df_1 = vis_df.groupby(by=["Line"]).min().reset_index()
            vis_df_1 = vis_df_1.sort_values(by=["Y"], ascending=True)
        answer_1 = list(vis_df_1["Line"])

        for file_nr in answer_data["fileNr"].unique():
            results = []
            answer_df = answer_data[(answer_data.blockType == 2) & (answer_data.visNr == i + 1) & (answer_data.fileNr == file_nr)]
            correct_answers = {
                1: list(answer_df[answer_df.questionNr == 1]["answer"])[0],
                2: list(answer_df[answer_df.questionNr == 2]["answer"])[0],
                3: list(answer_df[answer_df.questionNr == 3]["answer"])[0],
                4: list(answer_df[answer_df.questionNr == 4]["answer"])[0],
            }

            # Question 1
            results.append({
                "questionNr": 1,
                "note": "-",
                "correctAnswer": answer_1,
                "answer": correct_answers[1],
                "calculatedValue": answer_1 == correct_answers[1] or answer_1 == correct_answers[1][::-1],
                "valueType": "BOOL"
            })
            value_1_2 = min(len([j for j in range(0, 4) if answer_1[j] is not correct_answers[1][j]]),
                            len([j for j in range(0, 4) if answer_1[j] is not correct_answers[1][-j]]))
            results.append({
                "questionNr": 1,
                "note": "# errors in order",
                "questionType": QUESTION_TYPES_CALCULATED[3],
                "correctAnswer": len(answer_1),
                "calculatedValue": value_1_2,
                "valueType": "NUMERIC"
            })
            # Question 2
            vis_df_2 = vis_df[(vis_df.Line == correct_answers[1][0]) & (vis_df.X == x_value)]
            answer_2 = list(vis_df_2["Y"])[0]
            results.append({
                "questionNr": 2,
                "note": "-",
                "correctAnswer": answer_2,
                "answer": correct_answers[2],
                "calculatedValue": abs(answer_2 - correct_answers[2]),
                "valueType": "NUMERIC"
            })
            # Question 3
            vis_df_3 = vis_df[vis_df.Line.isin(answer_1[0:2])]
            vis_df_3 = vis_df_3.groupby(by=["Line"]).mean().reset_index()
            answer_3 = list(vis_df_3.sort_values(by="Y")["Line"])[0]
            results.append({
                "questionNr": 3,
                "note": "-",
                "correctAnswer": answer_3,
                "answer": correct_answers[3],
                "calculatedValue": answer_3 is correct_answers[3],
                "valueType": "BOOL"
            })
            # Question 4
            vis_df_4 = vis_df[vis_df.Line.isin(answer_1[2:4])]
            vis_df_4 = vis_df_4.groupby(by=["Line"])
            answer_4 = (list(vis_df_4.min().reset_index().sort_values(by=["Y"], ascending=True)["Y"])[0],
                        list(vis_df_4.max().reset_index().sort_values(by=["Y"], ascending=False)["Y"])[0])
            results.append({
                "questionNr": 4,
                "note": "-",
                "correctAnswer": answer_4,
                "answer": correct_answers[4],
                "calculatedValue": answer_4[0] is correct_answers[4][0] and answer_4[1] is correct_answers[4][1],
                "valueType": "BOOL"
            })
            results.append({
                "questionNr": 4,
                "note": "min",
                "questionType": QUESTION_TYPES_CALCULATED[0],
                "correctAnswer": answer_4[0],
                "answer": correct_answers[4][0],
                "calculatedValue": abs(answer_4[0] - correct_answers[4][0]),
                "valueType": "NUMERIC"
            })
            results.append({
                "questionNr": 4,
                "note": "max",
                "questionType": QUESTION_TYPES_CALCULATED[0],
                "correctAnswer": answer_4[1],
                "answer": correct_answers[4][1],
                "calculatedValue": abs(answer_4[1] - correct_answers[4][1]),
                "valueType": "NUMERIC"
            })
            results.append({
                "questionNr": 4,
                "note": "delta",
                "questionType": "-",
                "correctAnswer": answer_4[1] - answer_4[0],
                "answer": correct_answers[4][1] - correct_answers[4][0],
                "calculatedValue": abs((answer_4[1] - answer_4[0]) - (correct_answers[4][1] - correct_answers[4][0])),
                "valueType": "NUMERIC"
            })

            q_desc = QuestionDesc(file_nr, list(answer_df["background"])[0], list(answer_df["backgroundOrder"])[0], list(answer_df["sessionOrder"])[0], list(answer_df["blockNr"])[0], 2, i + 1)
            solution[q_desc] = results

    return solution


def calc_values_block3(vis_data: pd.DataFrame, answer_data: pd.DataFrame) -> Dict[QuestionDesc, List[Dict]]:
    question_values3 = [
        ["h", "l"],
        ["h", "l"],
        ["h", "l"],
        ["l", "h"],
        ["l", "h"],
        ["l", "h"],
        ["l", "h"],
        ["h", "l"],
    ]

    solution = {}
    for i, (line_select, extremum) in enumerate(question_values3):
        vis_df = vis_data[(vis_data.blockType == 3) & (vis_data.visNr == i + 1)]
        y_deltas = []
        for x in range(1, 36):
            vis_df_1 = vis_df[vis_df.X == x]
            ys = [
                int(vis_df_1[vis_df_1.Line == 0]["Y"]),
                int(vis_df_1[vis_df_1.Line == 1]["Y"]),
                int(vis_df_1[vis_df_1.Line == 2]["Y"]),
                int(vis_df_1[vis_df_1.Line == 3]["Y"])
            ]
            y_dist_values = [abs(ys[j1] - ys[j2]) for j1, j2 in itertools.combinations(list(range(0, 4)), 2)]
            y_deltas.append(sum(y_dist_values))
        answer_1 = y_deltas.index(min(y_deltas)) + 1

        for file_nr in answer_data["fileNr"].unique():
            results = []
            answer_df = answer_data[(answer_data.blockType == 3) & (answer_data.visNr == i + 1) & (answer_data.fileNr == file_nr)]
            correct_answers = {
                1: list(answer_df[answer_df.questionNr == 1]["answer"])[0],
                2: list(answer_df[answer_df.questionNr == 2]["answer"])[0],
                3: list(answer_df[answer_df.questionNr == 3]["answer"])[0],
                4: list(answer_df[answer_df.questionNr == 4]["answer"])[0],
            }

            # Question 1
            results.append({
                "questionNr": 1,
                "note": "-",
                "correctAnswer": answer_1,
                "answer": correct_answers[1],
                "calculatedValue": answer_1 is correct_answers[1],
                "valueType": "BOOL"
            })
            # Question 2
            vis_df_2 = vis_df[vis_df.X == correct_answers[1]]
            vis_df_2 = vis_df_2.groupby(by=["Line"])
            min_max_on_x = (list(vis_df_2.min().reset_index().sort_values(by=["Y"], ascending=True)["Y"])[0],
                            list(vis_df_2.max().reset_index().sort_values(by=["Y"], ascending=False)["Y"])[0])
            answer_2 = min_max_on_x[1] - min_max_on_x[0]
            results.append({
                "questionNr": 2,
                "note": "-",
                "correctAnswer": answer_2,
                "answer": correct_answers[2],
                "calculatedValue": abs(answer_2 - correct_answers[2]),
                "valueType": "NUMERIC"
            })
            # Question 3
            vis_df_3 = vis_df[vis_df.X == correct_answers[1]]
            if line_select == "h":
                selected_line = list(vis_df_3[vis_df_3.Y == min_max_on_x[1]]["Line"])[0]
            else:
                selected_line = list(vis_df_3[vis_df_3.Y == min_max_on_x[0]]["Line"])[0]
            vis_df_3 = vis_df[vis_df.Line == selected_line]
            if extremum == "h":
                answer_3 = vis_df_3["Y"].max()
            else:
                answer_3 = vis_df_3["Y"].min()
            results.append({
                "questionNr": 3,
                "note": "-",
                "correctAnswer": answer_3,
                "answer": correct_answers[3],
                "calculatedValue": abs(answer_3 - correct_answers[3]),
                "valueType": "NUMERIC"
            })
            # Question 4
            answer_4 = vis_df["Y"].mean()
            results.append({
                "questionNr": 4,
                "note": "-",
                "correctAnswer": answer_3 < answer_4,
                "answer": correct_answers[4],
                "calculatedValue": (answer_3 < answer_4) == correct_answers[4],
                "valueType": "BOOL"
            })

            q_desc = QuestionDesc(file_nr, list(answer_df["background"])[0], list(answer_df["backgroundOrder"])[0], list(answer_df["sessionOrder"])[0], list(answer_df["blockNr"])[0], 3, i + 1)
            solution[q_desc] = results

    return solution


def calc_time_values(reload_csv: bool = False) -> pd.DataFrame:
    os.makedirs(TEMPORARY_OUTPUT_PATH, exist_ok=True)
    file_path = f"{TEMPORARY_OUTPUT_PATH}/S2-derived-times-data.csv"
    if os.path.isfile(file_path) and not reload_csv:
        df = pd.read_csv(file_path)
        df["calculatedValue"] = df.apply(lambda row: data_string_converter(row["calculatedValue"]), axis=1)
        df["trimmedValue"] = df.apply(lambda row: data_string_converter(row["trimmedValue"]), axis=1)
        return df

    times_data = get_time_data()

    study_order = {}
    with open(STUDY_ORDER_INPUT_PATH, newline="") as json_file:
        study_order = json.load(json_file)

    output = {"fileNr": [], "background": [], "focus": [], "sessionOrder": [], "backgroundOrder": [], "blockType": [], "visNr": [],
              "questionType": [], "questionNr": [], "valueDescription": [],
              "calculatedValue": [], "valueType": []}
    for file_nr, file_content in times_data.items():
        for i, ((block_nr, question_nr, background), rows) in enumerate(file_content.items()):
            block_type = int((block_nr - 1) / 8) + 1
            vis_nr = (block_nr - 1) % 8 + 1
            session_order = (int(i / 4) + (0 if file_nr % 2 is 1 else 12)) % 24 + 1
            background_order = int(i / 4) % 12 + 1

            t_start = datetime.datetime.strptime(f"{rows[0][0]} {rows[0][1]}", "%Y-%m-%d %H:%M:%S.%f")
            t_first_toggle = datetime.datetime.strptime(f"{rows[1][0]} {rows[1][1]}", "%Y-%m-%d %H:%M:%S.%f")
            answer_indices = find_indices(rows, "Answering", 3)
            t_answer = datetime.datetime.strptime(f"{rows[answer_indices[0]][0]} {rows[answer_indices[0]][1]}", "%Y-%m-%d %H:%M:%S.%f")
            t_finished_answer = datetime.datetime.strptime(f"{rows[-1][0]} {rows[-1][1]}", "%Y-%m-%d %H:%M:%S.%f")

            toggle_counter = -1
            for i in range(0, answer_indices[0]):
                if "Toggle" in rows[i][3]:
                    toggle_counter += 1
            toggle_counter_answering = 0
            for i in range(answer_indices[0], len(rows)):
                if "Toggle" in rows[i][3]:
                    toggle_counter_answering += 1

            end_state = "Visualization" if toggle_counter % 2 == 0 else "Question"
            state_dur = {0: datetime.timedelta(0), 1: datetime.timedelta(0)}  # 0 -> vis, 1 -> question
            toggle_state, last_index = 0, 1
            for i in range(2, answer_indices[0] + 1):
                t0 = datetime.datetime.strptime(f"{rows[last_index][0]} {rows[last_index][1]}", "%Y-%m-%d %H:%M:%S.%f")
                t1 = datetime.datetime.strptime(f"{rows[i][0]} {rows[i][1]}", "%Y-%m-%d %H:%M:%S.%f")
                state_dur[toggle_state] += t1 - t0
                if "Toggle" in rows[i][3]:
                    toggle_state = (toggle_state + 1) % 2
                last_index = i

            answer_state_dur = {0: datetime.timedelta(0), 1: datetime.timedelta(0)}  # 0 -> vis, 1 -> question
            toggle_state, last_index = toggle_state, answer_indices[0]
            for i in range(answer_indices[0] + 1, len(rows)):
                t0 = datetime.datetime.strptime(f"{rows[last_index][0]} {rows[last_index][1]}", "%Y-%m-%d %H:%M:%S.%f")
                t1 = datetime.datetime.strptime(f"{rows[i][0]} {rows[i][1]}", "%Y-%m-%d %H:%M:%S.%f")
                answer_state_dur[toggle_state] += t1 - t0
                if "Toggle" in rows[i][3]:
                    toggle_state = (toggle_state + 1) % 2
                last_index = i
            answer_end_state = "Visualization" if (toggle_counter + toggle_counter_answering) % 2 == 0 else "Question"

            results = [
                ("StartDuration", (t_first_toggle - t_start).total_seconds(), "TIME"),
                ("SolvingDuration", (t_answer - t_first_toggle).total_seconds(), "TIME"),
                ("AnswerDuration", (t_finished_answer - t_answer).total_seconds(), "TIME"),
                ("CompleteDuration", (t_finished_answer - t_start).total_seconds(), "TIME"),
                ("Solving-VisDuration", state_dur[0].total_seconds(), "TIME"),
                ("Solving-QuestionDuration", state_dur[1].total_seconds(), "TIME"),
                ("Solving-EndState", end_state, "STRING"),
                ("Solving-#Toggles", toggle_counter, "TOGGLE"),
                ("Answering-VisDuration", answer_state_dur[0].total_seconds(), "TIME"),
                ("Answering-QuestionDuration", answer_state_dur[1].total_seconds(), "TIME"),
                ("Answering-EndState", answer_end_state, "STRING"),
                ("Answering-#Toggles", toggle_counter_answering, "TOGGLE")
            ]
            for (value_desc, value, value_type) in results:
                output["fileNr"].append(file_nr)
                output["background"].append(background)
                output["focus"].append(study_order[f"{file_nr}-{background}"][f"{block_nr:03d}"]["withFocusChange"])
                output["sessionOrder"].append(session_order)
                output["backgroundOrder"].append(background_order)
                output["blockType"].append(block_type)
                output["visNr"].append(vis_nr)
                output["questionType"].append(QUESTION_TYPE_MAP[block_type][question_nr])
                output["questionNr"].append(question_nr)
                output["valueDescription"].append(value_desc)
                output["calculatedValue"].append(value)
                output["valueType"].append(value_type)

    # Trim data
    def apply_trim_values(row, mean_df, error_df):
        if row["valueType"] == "STRING":
            return row["calculatedValue"]

        new_value = float(mean_df[mean_df.questionType == row["questionType"]]["calculatedValue"] +
                          2 * error_df[error_df.questionType == row["questionType"]]["calculatedValue"])
        return row["calculatedValue"] if row["calculatedValue"] <= new_value else new_value

    output_df = pd.DataFrame(output)
    tmp_df = output_df[output_df.valueType != "STRING"]
    tmp_df["calculatedValue"] = tmp_df["calculatedValue"].astype("float64")
    value_mean = tmp_df.groupby(by=["questionType"]).mean().reset_index()
    value_std = tmp_df.groupby(by=["questionType"]).std().reset_index()
    output_df["trimmedValue"] = output_df.apply(lambda row: apply_trim_values(row, value_mean, value_std), axis=1)
    output_df = convert_order_values(output_df, "sessionOrder")
    output_df = convert_order_values(output_df, "backgroundOrder")

    output_df.to_csv(file_path)
    return output_df


e = calc_values()
t = calc_time_values()
